import os
import sys
import time
import signal
import shutil
import psutil
sys.path.insert(0, "LCD_display")
from LCD_display import lcddriver
#import lcddriver
import subprocess
from subprocess import Popen, PIPE, DEVNULL, STDOUT, check_call
from random import randrange

# Global variables to store program state
tf_pid = -1
tf_ready = False

# Load LCD display driver
display = lcddriver.lcd()

# Signal handling
def signal_handler_SIGUSR2(sig, frame):
	print("> Signal 'SIGUSR2' received\n\n")
	# Set TF state to 'READY'
	global tf_ready
	tf_ready = True
	# Clear LCD display
	display.lcd_clear()
	
def signal_handler_SIGINT(sig, frame):
	print("\n\n> Signal 'SIGINT' received\n\n")
	print("Killing child processes\n\n")
	# Kill 'Object_detection_image.py' process
	kill_process = subprocess.Popen(["kill", "-s", "SIGKILL", str(tf_pid)], stdout=PIPE, stderr=PIPE)
	# Display farewell message
	print("Goodbye!\n\n")
	display.lcd_clear()
	display.lcd_display_string(" !! ABORTING !! ", 1)
	time.sleep(2)
	display.lcd_clear()
	display.lcd_display_string("    GOODBYE!    ", 1)
	time.sleep(0.5)
	display.lcd_clear()
	display.lcd_display_string("    GOODBYE!    ", 2)
	time.sleep(0.5)
	display.lcd_clear()
	display.lcd_display_string("    GOODBYE!    ", 1)
	time.sleep(0.5)
	display.lcd_clear()
	display.lcd_display_string("    GOODBYE!    ", 2)
	time.sleep(0.5)
	display.lcd_clear() # Leave LCD display cleaned up
	# Finish program
	print("Quitting!\n\n")
	exit(0) 
	
signal.signal(signal.SIGUSR2, signal_handler_SIGUSR2)
signal.signal(signal.SIGINT, signal_handler_SIGINT)

# Function definitions	
def file_size(filename):
	if os.path.isfile(filename):
		statinfo = os.stat(filename)
		return statinfo.st_size
	else: 
		return 0
		
def file_exists(filename):
	return os.path.isfile(filename)

def clear_LCD_display():
	display.lcd_clear()

def print_to_LCD_display(display, text = '', num_line = 1, display_time = 6, sticky = 0, num_cols = 16):
	# 'sticky' is only for short strings (len <= 16)
	if(len(text) > num_cols):
		timeout = time.time() + display_time # Set message timeout
		while True:
			if time.time() > timeout:
				display.lcd_clear() # Clear LCD display
				break
			else: 
				display.lcd_display_string(text[:num_cols],num_line)
				time.sleep(1)
				for i in range(len(text) - num_cols + 1):
					text_to_print = text[i:i+num_cols]
					display.lcd_display_string(text_to_print,num_line)
					time.sleep(0.2)
				time.sleep(1)
	else:
		display.lcd_display_string(text,num_line)
		if sticky == 0: 
			time.sleep(display_time)
			display.lcd_clear() # Clear LCD display
		
def print_to_LCD_display_both_short_short(display, text1 = "", text2 = "", display_time = 4):
	display.lcd_clear() # Clear LCD display
	display.lcd_display_string(text1, 1)
	display.lcd_display_string(text2, 2)
	time.sleep(display_time)
	display.lcd_clear()
	
def print_to_LCD_display_both_long_short(display, text1 = "", text2 = "", display_time = 6):
	display.lcd_clear() # Clear LCD display
	display.lcd_display_string(text2, 2)
	print_to_LCD_display(display, text1, 1, display_time)
	# previous function call clears LCD display after 'display_time'
	pass # 'close' function to wrap it
		
def print_to_LCD_display_both_short_long(display, text1 = "", text2 = "", display_time = 6):
	display.lcd_clear() # Clear LCD display
	display.lcd_display_string(text1, 1)
	print_to_LCD_display(display, text2, 2, display_time)
	# previous function call clears LCD display after 'display_time'
	pass # 'close' function to wrap it
	
def print_to_LCD_display_timer(display, total_sec = 99, seconds = 0, num_cols = 16):
	display.lcd_clear() # Clear LCD display
	if total_sec != 9 and total_sec-seconds == 9: display.lcd_clear()
	if total_sec != 99 and total_sec-seconds == 99: display.lcd_clear()
	adjust = int((seconds*num_cols)/total_sec)
	display.lcd_display_string("#"*adjust, 1)
	display.lcd_display_string("{}s".format(total_sec-seconds), 2)
	
def add_zeros_to_iterator(i):
	if i < 10000:
		n_zeros = 4 - len(str(i))
		formatted = "0"*n_zeros + str(i)
		return formatted # TYPE: String
	else:
		return str(i)
	
def get_next_record_iterator():
	lastfileHDD 	= subprocess.Popen("ls -t HDD/ | grep 'doc_' | head -n 1", 		shell=True, stdout=subprocess.PIPE).communicate()[0].decode('utf-8').strip()
	if not lastfileHDD: iteratorHDD = 0
	else: iteratorHDD = int(lastfileHDD[4:-4])
	
	lastfilePrinted = subprocess.Popen("ls -t printed/ | grep 'doc_' | head -n 1", 	shell=True, stdout=subprocess.PIPE).communicate()[0].decode('utf-8').strip()
	if not lastfilePrinted: iteratorPrinted = 0
	else: iteratorPrinted = int(lastfilePrinted[4:-4])
	
	if iteratorHDD > iteratorPrinted:
		iterator = iteratorHDD
	else: 
		iterator = iteratorPrinted
	
	return iterator + 1
    
def get_last_HDD_iterator():
	lastfile = subprocess.Popen("ls -t HDD/ | grep 'doc_' | head -n 1", shell=True, stdout=subprocess.PIPE).communicate()[0].decode('utf-8').strip()
	if not lastfile: iterator = 0
	else: iterator = int(lastfile[4:-4])
	return iterator

def get_random_existing_printed_document():
	firstfile = subprocess.Popen("ls -tr printed/ | grep 'doc_' | head -n 1", shell=True, stdout=subprocess.PIPE).communicate()[0].decode('utf-8').strip()
	lastfile = subprocess.Popen("ls -t printed/ | grep 'doc_' | head -n 1",  shell=True, stdout=subprocess.PIPE).communicate()[0].decode('utf-8').strip()
	return randrange(int(firstfile[4:-4]), int(lastfile[4:-4]))

def get_output_classes_sentence(output_classes):
	if not output_classes: sentence = "nothing" # no objects detected
	else:
		classes = output_classes.split("\n")
		resdict = {}
		for aclass in classes:
			if aclass in resdict: 
				resdict[aclass] += 1 # increment counter of object 'aclass'
			else: 
				resdict[aclass] = 1 # new objected detected
				
		# create sentence
		sentence = ""
		first_it = True
		for key, value in sorted(resdict.items(), key=lambda k: k[1]):
			if first_it:
				sentence += "{0}X {1}".format(value, key)
				first_it = False
			else:
				sentence += ", {0}X {1}".format(value, key)
	# return generated sentence
	return sentence
	

# MAIN PROGRAM

# TensorFlow object_detection process initialization (without waiting)
os.chdir('/home/alvaro/tensorflow/models/research/object_detection')
tf_process = subprocess.Popen(["/usr/bin/python3", "-W", "ignore", "Object_detection_image.py", "--simple", "--ppid", str(os.getpid())], stdout=PIPE, stderr=PIPE)
tf_pid = tf_process.pid

# Set working directory
os.chdir('/home/alvaro'); 

# Welcome message shell
print("\n\n")
print("##################################################################################\n\n")
print("###################### MÀQUINA DE RECORDS by Àlvaro Cànovas ######################\n\n")
print("##################################################################################\n\n")
time.sleep(1)
print("> Stop script with 'SIGINT' [Ctrl+C]\n\n")
print("> Starting in directory", os.getcwd(), "with PID =", os.getpid(),"\n\n")
print("> The TensorFlow API process has been started with PID =", tf_pid, "\n\n")
print("__________________________________________________________________________________\n\n\n")

# Welcome message LCD
print_to_LCD_display(display, "MAQUINA DE RECORDS by Alvaro Canovas")
print_to_LCD_display(display, "    Welcome!    ", 2, 2)
print_to_LCD_display_both_short_short(display, "Stop script with", "SIGINT [Ctrl+C]")
#print_to_LCD_display(display, "Starting in directory {}".format(os.getcwd()))
#print_to_LCD_display(display, "The TensorFlow API process has been started with PID = {}".format(tf_pid))

# Wait until TensorFlow resources are ready
msg = "Getting ready all the TensorFlow resources"
print(msg, "| Waiting 'SIGUSR2'\n\n")
print_to_LCD_display(display, msg)
print_to_LCD_display(display, "Getting ready...", 2, 0, 1) # Screen sticky until SIGUSR2

while tf_ready == False: # Polling
	time.sleep(1)
tf_ready = False # set state to False for next polling zone

# Get next image id iterator 
i = get_next_record_iterator()

# Ready to start capturing images
print("Ready to start capturing images with 'record_id' = {}\n\n".format(i))
print_to_LCD_display_both_short_short(display, "Starting  with", "record-id: {}".format(i))

print("__________________________________________________________________________________\n\n\n")

first_it = True # Firs iterations will always be 'CAPTURE' mode

# Main loop
while True:
	# Ready to start iteration
	print("Ready!\n\n")
	clear_LCD_display()
	print_to_LCD_display(display, "     Ready!     ", 1, 2)
	
	# Select working mode randomly [10%]
	mode = randrange(0,9)
	
	if mode >= 1 or first_it == True: # Capture 'record' working mode
		first_it = False # curr_iteration id > FIRST
		msgs = ["WORKING MODE:", "CAPTURE"] 
		print(msgs[0], msgs[1], "\n\n") # only terminal log
		
		# Generate filename & 'Click!'
		filename = "record.jpg";
		print("Click\n\n")
		print_to_LCD_display(display, "   CAPTURANT!   ", 1, 2)
		
		# Capture image
		subprocess.Popen(["raspistill", "-n", "-t", "1", "-w", "600", "-h", "925", "-o", filename], stdout=PIPE, stderr=PIPE).wait()
		filesize = file_size(filename)/1024
		print("Capturing '{0}' [{1:.2f} KB]".format(filename, filesize), "\n\n")
		print_to_LCD_display_both_short_short(display, "Record_{}".format(add_zeros_to_iterator(i)), "[{0:.2f} KB]".format(filesize))
		
		if not file_exists(filename): break # Errors found 
		
		# Generate pdf document in this directory via './aux/generate_document.php'
		docname = "doc_{}.pdf".format(i)
		pdf_process = subprocess.Popen(["/usr/bin/php", "/home/alvaro/aux/generate_document.php", add_zeros_to_iterator(i), filename, docname], stdout=PIPE, stderr=PIPE)
		pdf_stdout = pdf_process.communicate()[0]
		docsize = file_size(docname)/1024
		print("Generating document '{0}' [{1:.2f} KB]\n\n".format(docname, docsize))
		time.sleep(0.5) # time indent terminal logs
		
		if pdf_stdout.decode('utf-8').strip() != "Done": break # Errors found
		
		# Send signal to 'Object_detection_image.py' in order to start analyzing the objects in the image
		t_ini = time.time() # Measure elapsed time
		os.kill(tf_pid, signal.SIGUSR1) # Signal SIGUSR1
		print("Sending signal 'SIGUSR1' to the TensorFlow API to start object detecting\n\n") # only terminal log
		time.sleep(1)
		
		# Wait until TensorFlow has detected the objects and written the output to '~/output_classes.txt'
		msgs = ["TensorFlow API", "Object Detection"];
		print(msgs[0], "running", msgs[1], "| Waiting 'SIGUSR2'\n\n")
		print_to_LCD_display_both_short_short(display, " {} ".format(msgs[0]), msgs[1], 6)
		print_to_LCD_display(display, "  Analyzing...  ", 2, 0, 1) # Screen sticky
		while tf_ready == False: # Polling
			check_tf_proc = psutil.Process(tf_pid)
			print(check_tf_proc.status(), end=", ", flush=True)
			time.sleep(1)
		elapsed_time = time.time() - t_ini # 'Object_detection_image.py' detection function has finished
		tf_ready = False # set state to False for next polling zone
		
		# Show output generated by the TensorFlow API
		file = open("/home/alvaro/output_classes.txt", "r")
		file.seek(12)
		msgs = ["Detected:", get_output_classes_sentence(file.read())]
		print("After {1:.2f}s -> {0} {2}\n\n".format(msgs[0], elapsed_time, msgs[1]))
		print_to_LCD_display_both_short_long(display, msgs[0], msgs[1], 8)
		
		# Remove captured image once the document is created and TensorFlow has analyzed it
		os.remove(filename)
		
		# Find out where the result document should be moved to
		total, used, free = shutil.disk_usage("HDD/")
		freespace = free/2**20 # MB
		print("HDD free space: {0:.2f} MB\n\n".format(freespace))
		print_to_LCD_display_both_short_short(display, "Espai Memoria:", "{0:.2f} MB".format(freespace))

		# Print document via CUPS
		if freespace < 128: # MB
			#printer = "[WIFI]_Brother_DCP-1610W"; # set printer
			printer = "EPSON XP-5100 Series"; # set printer
			subprocess.Popen(["/usr/bin/lp", docname], stdout=PIPE, stderr=PIPE).wait() # Print file
			print("Printing captured image via", printer, "\n\n")
			print_to_LCD_display_both_short_short(display, "Externalitzant", "Record_{}".format(add_zeros_to_iterator(i)))
			# specify destination folder
			outdir = "printed/"  
			
		# Move document to HDD
		else:
			print("Moving document to the Hard Drive\n\n")
			print_to_LCD_display_both_short_short(display, "Guardant a la", "memoria")
			# specify destination folder
			outdir = "HDD/"
		
		# move document to specified folder [HDD/ or printed/]
		mv_process = shutil.move(docname, outdir + docname)
			
		# Increment 'record_id' iterator
		i += 1
		
	else: # Retrieve 'record' working mode
		msgs = ["WORKING MODE:", "RETRIEVE"] 
		print(msgs[0], msgs[1], "\n\n") # only terminal log
		
		# Try to retrieve randomly selected document
		doc_id = get_random_existing_printed_document()
		print("Trying to retrieve document with 'record-id' = {}\n\n".format(doc_id))
		print_to_LCD_display_both_short_short(display, "Desig de", "Record_{}".format(add_zeros_to_iterator(doc_id)))
		
		# Display message: Place document inside the scanner [15s]
		print("Place document inside the scanner | Waiting...\n\n")
		print_to_LCD_display_both_long_short(display, "Esperant input memoria externa:", "Record_{}".format(add_zeros_to_iterator(doc_id)))
		wait_time = 15
		print("-"*wait_time, "\n")
		for k in range(wait_time):
			print("#", sep='', end='', flush=True)
			print_to_LCD_display_timer(display, wait_time, k)
			time.sleep(1) # 1 second
		print("\n\n")
		
		# Scan document as TIFF image
		print("Scanning...\n\n")
		print_to_LCD_display(display, "  RECORDANT...  ", 1, 4, 1)
		filename = "record.jpg";
		subprocess.Popen("/usr/bin/scanimage -p --format=tiff > record.jpg", shell=True, stdout=PIPE, stderr=PIPE).wait()
		
		if not file_exists(filename): break # Errors found 
		
		# Send signal to 'Object_detection_image.py' in order to start analyzing the objects in the image
		t_ini = time.time() # Measure elapsed time
		os.kill(tf_pid, signal.SIGUSR1) # Signal SIGUSR1
		print("Sending signal 'SIGUSR1' to the TensorFlow API to start object detecting\n\n") # only terminal log
		time.sleep(1)
		
		# Wait until TensorFlow has detected the objects and written the output to '~/output_classes.txt'
		msgs = ["TensorFlow API", "Object Detection"];
		print(msgs[0], "running", msgs[1], "| Waiting 'SIGUSR2'\n\n")
		print_to_LCD_display_both_short_short(display, " {} ".format(msgs[0]), msgs[1], 6)
		print_to_LCD_display(display, "  Analyzing...  ", 2, 0, 1) # Screen sticky
		while tf_ready == False: # Polling
			check_tf_proc = psutil.Process(tf_pid)
			print(check_tf_proc.status(), end=", ", flush=True)
			time.sleep(1)
		elapsed_time = time.time() - t_ini # 'Object_detection_image.py' detection function has finished
		tf_ready = False # set state to False for next polling zone
		
		# Show output generated by the TensorFlow API
		file = open("/home/alvaro/output_classes.txt", "r")
		file.seek(12)
		msgs = ["Detected:", get_output_classes_sentence(file.read())]
		print("After {1:.2f}s -> {0} {2}\n\n".format(msgs[0], elapsed_time, msgs[1]))
		print_to_LCD_display_both_short_long(display, msgs[0], msgs[1], 8)
		
		# Remove scanned image once TensorFlow has analyzed it
		os.remove(filename)

	# Sleep
	sleep_time = randrange(2,180); # seconds
	
	print("I am going to sleep for {} seconds".format(sleep_time), "\n\n")
	
	print("-"*sleep_time, "\n")
	for k in range(sleep_time):
		print("#", sep='', end='', flush=True)
		print_to_LCD_display_timer(display, sleep_time, k)
		time.sleep(1) # 1 second
	print("\n\n\n________________________________________________________________________\n\n\n")
	

# Some error caused the loop to break
print("\n\n\n> ERROR!\n\n\n")
print_to_LCD_display(display, "   < ERROR! >  ", 1, 1)
print_to_LCD_display(display, "   < ERROR! >  ", 2, 1)
print_to_LCD_display(display, "   < ERROR! >  ", 1, 1)
print_to_LCD_display(display, "   < ERROR! >  ", 2, 1)
print_to_LCD_display(display, "   < ERROR! >  ", 1, 1)
print_to_LCD_display(display, "   < ERROR! >  ", 2, 1)
print_to_LCD_display(display, "   < ERROR! >  ", 1, 1)
print_to_LCD_display(display, "   < ERROR! >  ", 2, 1)
print_to_LCD_display(display, "   < ERROR! >  ", 1)

